"""Provides a class to interact with a subset of the Twitch API.

Specifically, the "new Twitch API", which they haven't put a version
number on yet.
"""

import requests
from twitchgamenotify.constants import (
    HTTP_200_OK,
    HTTP_401_UNAUTHORIZED,
    TWITCH_GAME_API_URL,
    TWITCH_STREAM_API_URL,
    TWITCH_TOKEN_API_URL,
    TWITCH_USER_API_URL,)


class FailedHttpRequest(Exception):
    """An exception raised when an HTTP request failed."""
    def __init__(self, message, http_status_code):
        """Record what the HTTP status code for the bad request was."""
        # Call the parent class __init__
        super().__init__(message)

        # Record the status code and message
        self.status_code = http_status_code
        self.message = message


class TwitchApi:
    """Interacts with the Twitch API."""
    def __init__(self, client_id, client_secret):
        """Set up authorization."""
        # Load in authentication details
        self.client_id = client_id
        self.client_secret = client_secret

        # Start a requests session
        self.session = requests.Session()

        # Get and set an access token
        self.obtain_access_token()

    def obtain_access_token(self):
        """Obtains and sets a fresh access token."""
        # Get the access token
        response = requests.post(
            TWITCH_TOKEN_API_URL
            + '?client_id='
            + self.client_id
            + '&client_secret='
            + self.client_secret
            + '&grant_type=client_credentials')

        try:
            # Make the the HTTP request was okay
            assert response.status_code == HTTP_200_OK
        except AssertionError:
            # The HTTP request wasn't okay
            message = (
                "An access token fetch failed with status code %s" %
                response.status_code)

            raise FailedHttpRequest(
                message=message,
                http_status_code=response.status_code,)

        # Set the access token
        access_token = response.json()['access_token']
        self.session.headers.update(
            {'Authorization': 'Bearer ' + access_token})

    def make_http_request(self, http_request_url):
        """Makes an HTTP request.

        This assumes that all incoming HTTP requests are GETs, which
        will be true for everything passed to this function within the
        scope of this program.

        If an access token is expired, another one is obtained.

        Args:
            http_request_url: A string containing the URL to make an
                HTTP request to.

        Returns:
            A requests.models.Response object containing the response to
            the successful HTTP request.

        Raises:
            FailedHttpRequest: The status code indicated the HTTP
                request was not successful.
        """
        # Make the request
        response = self.session.get(http_request_url)

        # If our access token has expired, get another one and retry the
        # request
        if response.status_code == HTTP_401_UNAUTHORIZED:
            # Get a new access token
            self.obtain_access_token()

            # Repeat the request
            response = self.session.get(http_request_url)

        try:
            # Make the the HTTP request was okay
            assert response.status_code == HTTP_200_OK
        except AssertionError:
            # The HTTP request wasn't okay
            message = (
                "An HTTP request to %s failed with status code %s" %
                (http_request_url, response.status_code,))

            raise FailedHttpRequest(
                message=message,
                http_status_code=response.status_code,)

        return response

    def get_online_stream_info(self, streamer_login_name):
        """Requests info about an online stream.

        Arg:
            streamer_login_name: A string specifying the streamer's user
                login name. E.g., moonmoon_ow.

        Returns:
            A dictionary of information about the queried stream
            including whether it's live, its title, and the current
            game's game ID. For example:

            {'live': True,
             'title': "Macie Jay Charm PogChamp",
             'game_id': '460630'}
        """
        # Make a request to the Twitch API
        response = self.make_http_request(
            TWITCH_STREAM_API_URL
            + '?user_login='
            + streamer_login_name)

        # Build up the info of this stream
        response_data = response.json()['data']

        if not response_data:
            # Stream is offline
            stream_info = dict(
                live=False,
                title="",
                game_id="",)
        else:
            stream_info = dict(
                live=True,
                title=response_data[0]['title'],
                game_id=response_data[0]['game_id'],)

        return stream_info

    def get_streamer_display_name(self, streamer_login_name):
        """Get a streamer's display name given their login name.

        Arg:
            streamer_login_name: A string specifying the streamer's user
                login name. E.g., moonmoon_ow.

        Returns:
            A string containing the streamer's display name.
        """
        # Make a request to the Twitch API
        response = self.make_http_request(
            TWITCH_USER_API_URL
            + '?login='
            + streamer_login_name)

        # Return the display name
        return response.json()['data'][0]['display_name']

    def get_game_title(self, game_id):
        """Get a game's title given its ID.

        Arg:
            game_id: A string containing a Twitch game ID.

        Returns:
            A string containing the game's title.
        """
        # Make a request to the Twitch API
        response = self.make_http_request(
            TWITCH_GAME_API_URL
            + '?id='
            + game_id)

        # Return the game title
        return response.json()['data'][0]['name']



"""Contains the main function."""

import atexit
import logging
import time
import threading
import sys
import notify2
import requests
from twitchgamenotify.cache import (
    load_cache,
    save_cache,)
from twitchgamenotify.constants import (
    STARTING_CONNECTION_FAILURE_RETRY_TIME,)
from twitchgamenotify.configuration import (
    ConfigFileNotFound,
    parse_config_file,
    parse_runtime_args,)
from twitchgamenotify.notifications import (
    process_notifications_wrapper,
    send_connection_error_notification,)
from twitchgamenotify.twitch_api import (
    TwitchApi,)
from twitchgamenotify.version import NAME


def main():
    """The main function."""
    # Get runtime arguments
    cli_args = parse_runtime_args()

    # Set up logger
    logging.basicConfig(
        format='%(levelname)s: %(message)s',
        level=cli_args.loglevel,)

    # Load cached static API data
    if not cli_args.no_caching:
        # Load the cache
        cache_dict = load_cache()

        # Create a hook to save the cache when exiting the program
        atexit.register(save_cache, cache_dict)

    # Read config file
    try:
        config_dict = parse_config_file()
    except ConfigFileNotFound:
        logging.error("Config file not found. Aborting.")
        sys.exit(1)

    # Set up the notifier
    if not cli_args.print_to_terminal:
        notify2.init(NAME)

    # Set up app indicator and run it in a separate thread
    if (not cli_args.one_shot
            or not cli_args.print_to_terminal
            or cli_args.no_app_indicator):
        # Import this here to GTK-incompatible machines are still
        # supported
        from twitchgamenotify.app_indicator import AppIndicator

        indicator = AppIndicator()

        # Kill the indicator when we quit
        atexit.register(indicator.stop)

        # Start the indicator in its own thread
        threading.Thread(target=indicator.start, daemon=True).start()

    # Connect to the API - keep retrying and be loud if there's a
    # connection error
    while True:
        try:
            twitch_api = TwitchApi(
                client_id=config_dict['twitch-api-client-id'],
                client_secret=config_dict['twitch-api-client-secret'],)

            # Successful connection
            break
        except requests.exceptions.ConnectionError:
            # Internet is probably down. Log an error and notify if we're
            # notifying
            send_connection_error_notification(
                send_dbus_notification=not cli_args.print_to_terminal)

            logging.info(
                "Retrying in %s seconds",
                STARTING_CONNECTION_FAILURE_RETRY_TIME,)

            # Wait a bit before retrying
            time.sleep(STARTING_CONNECTION_FAILURE_RETRY_TIME)

    # Set up arguments to give process_notifcations
    kwargs = dict(
        print_to_terminal=cli_args.print_to_terminal,
        streamers=config_dict['streamers'],
        twitch_api=twitch_api,)

    if not cli_args.no_caching:
        kwargs['names_cache'] = cache_dict

    if not cli_args.one_shot:
        # Remember what game a streamer was playing last so we don't
        # re-notify
        streamers_last_seen_playing_dict = (
            {streamer: "" for streamer in config_dict['streamers'].keys()})

        kwargs['streamers_previous_game'] = streamers_last_seen_playing_dict

    # Query (and possibly notify) only once or periodically
    if cli_args.one_shot:
        process_notifications_wrapper(**kwargs)
    else:
        # Loop until we hit a keyboard interrupt
        try:
            while True:
                # Process any notifications
                threading.Thread(
                    target=process_notifications_wrapper,
                    kwargs=kwargs,).start()

                # Wait before querying again
                time.sleep(config_dict['query-period'])
        except KeyboardInterrupt:
            logging.info("Exitting %s", NAME)



"""Functions for caching static API data."""

import json
import os
from twitchgamenotify.constants import (
    CACHE_FILE_NAME,
    PROJECT_CONFIG_HOME,)


def load_cache():
    """Loads a cache, or creates one if it doesn't exit.

    The cache is located at
    $XDG_CONFIG_HOME/twitch-game-notify/cache.json (XDG_CONFIG_HOME
    defaults to $HOME/.config).

    Returns:
        A dictionary containing static API data. For example:
    """
    # Build the path to the cache JSON file
    cache_path = os.path.join(PROJECT_CONFIG_HOME, CACHE_FILE_NAME)

    # If a cache doesn't exist, return an empty dictionary as the new
    # cache
    if not os.path.exists(cache_path):
        return {'games': {}, 'streamers': {}}

    # Load the cache file
    with open(cache_path, 'r') as cache_file:
        return json.load(cache_file)

def save_cache(cache_dictionary):
    """Saves a cache as JSON.

    The cache is located at
    $XDG_CONFIG_HOME/twitch-game-notify/cache.json (XDG_CONFIG_HOME
    defaults to $HOME/.config).

    Arg:
        cache_dictionary: A dictionary containing cache information to
            be serialized to JSON.
    """
    # Get the path for the cache
    cache_path = os.path.join(PROJECT_CONFIG_HOME, CACHE_FILE_NAME)

    # Make sure the directory paths to the cache file exist
    os.makedirs(PROJECT_CONFIG_HOME, exist_ok=True)

    with open(cache_path, 'w') as cache_file:
        cache_file.write(json.dumps(cache_dictionary))



"""Stores constants for all modules."""

import logging
import os


# Base of the repository/project
PROJECT_BASE_DIR = (
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

# Path to the app indicator icon
APP_INDICATOR_SVG_PATH = os.path.join(
    PROJECT_BASE_DIR,
    'static/',
    'twitch.svg',)

# Base of XDG config files
try:
    PROJECT_CONFIG_HOME = os.path.join(
        os.environ['XDG_CONFIG_HOME'],
        'twitch-game-notify')
except KeyError:
    PROJECT_CONFIG_HOME = os.path.join(
        os.environ['HOME'],
        '.config/',
        'twitch-game-notify',)

# Config file names
CONFIG_FILE_NAME = 'config.yaml'
CACHE_FILE_NAME = 'cache.json'

# URL for example config file on Github
EXAMPLE_CONFIG_FILE_URL = (
    'https://raw.githubusercontent.com/mwiens91/twitch-game-notify/master/config.yaml.example')

# Loglevel CLI options
CRITICAL = 'critical'
ERROR = 'error'
WARNING = 'warning'
INFO = 'info'
DEBUG = 'debug'

LOGLEVEL_CHOICES = [CRITICAL, ERROR, WARNING, INFO, DEBUG]
LOGLEVEL_DICT = {CRITICAL: logging.CRITICAL,
                 ERROR: logging.ERROR,
                 WARNING: logging.WARNING,
                 INFO: logging.INFO,
                 DEBUG: logging.DEBUG,}

# Twitch API URLs
TWITCH_BASE_API_URL = 'https://api.twitch.tv/helix'
TWITCH_STREAM_API_URL = TWITCH_BASE_API_URL + '/streams'
TWITCH_USER_API_URL = TWITCH_BASE_API_URL + '/users'
TWITCH_GAME_API_URL = TWITCH_BASE_API_URL + '/games'
TWITCH_TOKEN_API_URL = 'https://id.twitch.tv/oauth2/token'

# HTTP status codes
HTTP_200_OK = 200
HTTP_401_UNAUTHORIZED = 401

# How many seconds to wait when initially no connection
STARTING_CONNECTION_FAILURE_RETRY_TIME = 10



#!/usr/bin/env python3

from setuptools import setup
from twitchgamenotify.version import NAME, DESCRIPTION, VERSION


# Parse readme to include in PyPI page
with open('README.md') as f:
    long_description = f.read()

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION.capitalize(),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/mwiens91/twitch-game-notify',
    author='Matt Wiens',
    author_email='mwiens91@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3 :: Only',
    ],
    packages=['twitchgamenotify', 'static'],
    package_data={'static': ['twitch.svg']},
    entry_points={
        'console_scripts': ['twitch-game-notify = twitchgamenotify.main:main'],
    },
    python_requires='>=3',
    install_requires=[
        'dbus-python',
        'notify2',
        'PyGObject',
        'PyYAML',
        'requests',
    ],
)



"""Contains configuration related functions."""

import argparse
import os.path
import sys
import requests
import yaml
from twitchgamenotify.constants import (
    CONFIG_FILE_NAME,
    EXAMPLE_CONFIG_FILE_URL,
    LOGLEVEL_CHOICES,
    LOGLEVEL_DICT,
    PROJECT_BASE_DIR,
    PROJECT_CONFIG_HOME,
    WARNING,)
from twitchgamenotify.version import NAME, VERSION, DESCRIPTION


class ConfigFileNotFound(Exception):
    """Raised when a config file can't be found."""
    pass


class TranslateLogLevelAction(argparse.Action):
    """argparse action to translate loglevel to a number."""
    def __call__(self, parser, namespace, values, option_string=None):
        """Set a logging loglevel number."""
        setattr(namespace, self.dest, LOGLEVEL_DICT[values])


class PrintExampleConfigAction(argparse.Action):
    """argparse action to print example config file."""
    def __call__(self, parser, namespace, values, option_string=None):
        """Download and print example config file."""
        r = requests.get(EXAMPLE_CONFIG_FILE_URL)

        if r.status_code != 200:
            print(
                "ERROR: Failed to download example config file!",
                file=sys.stderr,)
            sys.exit(1)
        else:
            print(r.content.decode())
            sys.exit(0)


def find_config_file():
    """Find and return the path of a config file.

    The config file looked for is "config.yaml" and it is looked for at
    the base of the respository first (if you're running from source),
    and then in $XDG_CONFIG_HOME/twitch-game-notify/ (XDG_CONFIG_HOME
    defaults to $HOME/.config).

    Returns:
        A string containing the absolute path to the config file.

    Raises:
        ConfigFileNotFound: A config file couldn't be found.
    """
    # Check the base of the project
    config_path = os.path.join(PROJECT_BASE_DIR, CONFIG_FILE_NAME)

    if os.path.exists(config_path):
        return config_path

    # Check XDG_CONFIG_HOME
    config_path = os.path.join(PROJECT_CONFIG_HOME, CONFIG_FILE_NAME)

    if os.path.exists(config_path):
        return config_path

    # Couldn't find anything :thinking:
    raise ConfigFileNotFound


def parse_config_file():
    """Find and parse a config file.

    Raises:
        ConfigFileNotFound: A config file couldn't be found.
    """
    # Find the config file first
    config_path = find_config_file()

    # Now parse and return it - note that PyYAML doesn't come with any
    # schema validation, which might be desirable at some point
    with open(config_path, 'r') as config_file:
        return yaml.load(config_file)


def parse_runtime_args():
    """Parse runtime args using argparse.

    Returns:
        An object of type 'argparse.Namespace' containing the runtime
        arguments as attributes. See argparse documentation for more
        details.
    """
    parser = argparse.ArgumentParser(
        prog=NAME,
        description="%(prog)s - " + DESCRIPTION,)
    parser.add_argument(
        '-l', '--loglevel',
        default=LOGLEVEL_DICT[WARNING],
        choices=LOGLEVEL_CHOICES,
        action=TranslateLogLevelAction,
        help="how much to log",)
    parser.add_argument(
        '--one-shot',
        action='store_true',
        help="query once then exit")
    parser.add_argument(
        '--no-app-indicator',
        action='store_true',
        help="don't launch an app indicator")
    parser.add_argument(
        '--no-caching',
        action='store_true',
        help="don't cache static API data")
    parser.add_argument(
        '--print-config',
        nargs=0,
        action=PrintExampleConfigAction,
        help="downloads and prints example config file")
    parser.add_argument(
        '--print-to-terminal',
        action='store_true',
        help="print to terminal (doesn't connect to D-Bus)")
    parser.add_argument(
        '--version',
        action='version',
        version="%(prog)s " + VERSION)

    return parser.parse_args()
